/*  
    AUTHOR and COPYRIGHT(C)2009: Dan Dulberger, evenflcw@yahoo.com
    You may not redistribute this script, in original or modified form, without explicit consent from the author.
*/
/*
TODO:
-implement lastcommands array in commandList, commandName(command), and enough functions in commandList not to refer to any of the global arrays in interface module.
-create backup copy of config upon load. its last known good config.
-maybe fuse all lists into one? this would allow us to print all commands easily. would perhaps perform slower though.
-Add tab with listbox for viewing and editing blocked commands.
-Save out command list. Use this to validate commands. Also allow users to tag commands as valid commands.
-Fix resizing.
-Analyze mouse for proper tick/untick in listbox
-Draw the gridlines of the listbox or implement own listbox.
-Allow for multiple arguments in commands.
-Allow for arguments, ie 'ItemVisibility 4'. This would also give support to commands with spaces in them. Necessary?
-Dynamic mode, processes all commands except excluded and blacklisted commands. Risky!

TODO for version2.
-new gui. three lists. enabled commands - all commands - blocked commands. and arrows inbetween. as you select a command in either list. the others update to match the selection.
-add pluginservice. either by incorporating the plugin or reimplementing it.



*/

@version 2.7
@warnings
@script master
@name "ef MultiService"
@define NAME "ef MultiService with Last Command"
@define SERVERNAME "ef_MultiService"
@define VERSION 1.99


///
///		COMMAND LIST - START
///

@sequence {
 COMMAND_NAME,
 COMMAND_STATE
}
// @sequence { //new states for joined list
 // COMMAND_STATE_DISABLED	=0, 		//MultiService doesn't do anything
 // COMMAND_STATE_ENABLED	=2,			//MultiService provides multiple selection service
 // COMMAND_STATE_BLOCKED	=4,			//MultiService ignores, LastCommand ignores
 // COMMAND_STATE_NOARGUMENTS	=8		//Issues the command without arguments, usefull 
// }
@sequence {
 COMMAND_STATE_INACTIVE=0,
 COMMAND_STATE_ACTIVE,
 COMMAND_STATE_BLOCKED
}
listNewElement: data {
	element[COMMAND_NAME] = data[COMMAND_NAME];
	element[COMMAND_STATE] = data[COMMAND_STATE];
	return element;
}
listCompareElements: a, b {
	return	a[COMMAND_NAME] == b[COMMAND_NAME] ? 0 : 1; //return 0 if match
}
@define LIST_NEW_ELEMENT listNewElement
@define LIST_COMPARE_ELEMENTS listCompareElements


///
///		LIBLIST - START
///

/*
	List by Dan Dulberger
	
	To use:
	
	(1) Insert using:
		(at)insert "D:\\Development_Files\\Projects\\LightWave_Scripts\\Include\\list.lsi"
		

	Elements can be single variable or complex array.
	Elements must be edited in-place. Either directly by indexing the list, or via custom function (taking list element adress as arg).

	@sequence {
		ELEMENT_PROPERTY_NAME,	//
		ELEMENT_PROPERTY_STATE,
		...
	}
	listNewElement(data, ...) //creates a new element from the data provided.
	listAddElement(list, element) //list must be initialized, element must be valid. What about element being an array?
	listRemoveElementByIndex(list, index) //removes element at index, packs and truncs. To remove by id, use find() first.
	listFindElement(list, property, value) //searches through list for element with property matching value. returns array of indices.
	listCompareElements(a,b)
	
	copy = listCopyElement(element)
	
	one could possible implement unique ids so one would not be dependant on index. hashes or just assign i+1 for everynew item... but maxint.
	
	The list is fully implemented. But elements are simple varibles.
	To use own data structure for element one must define some case specific functions.
	LIST_NEW_ELEMENT
	LIST_COMPARE_ELEMENTS
	
*/
/*
Example implementation (WORKS!):

newListElement: data {
	element[1] = data[1];		//element can be jagged n-dim array
	element[2][1] = data[2];
	element[3][1] = data[3];
	return element;
}
listCompareElements: a,b {
	return a[2][1] == b[2][1];
}

(at)define LIST_NEW_ELEMENT newListElement
(at)define LIST_COMPARE_ELEMENTS listCompareElements
(at)insert "D:\\Development_Files\\Projects\\LightWave_Scripts\\Include\\list.lsi"
(at)define MYPROPERTY [2][1]

generic {
debug();
	tlist = newList(10);
	telmt = newListElement(@123,"abc","def"@);
	listAddUniqueElement(tlist, telmt);
	listAddUniqueElement(tlist, telmt);
	telmt = newListElement(@123,"abd","def"@);
	listAddUniqueElement(tlist, telmt);
	tlist[2]MYPROPERTY = "xxx";	//see define

*/


/**
	Creates a new list.
*/
listNew: size {
	list[size] = nil;
	return list;
}


/**
	Creates a new element.
	Should be overriden by case-specific define.
*/
_listNewElement: data {
	return data;
}


/**
	Compares two elements
	Should be overriden by case-specific define.
*/
_listCompareElements: a, b {
	if(a<b)
		return -1;
	if(a>b)
		return 1;
	return 0;	
}


/**
	Returns the index of a matching element, else 0;
	For simple elements.
*/
listFindElement: list, element { 
	for(i=1; i<=list.count(); i++) {
		if(LIST_COMPARE_ELEMENTS(list[i], element) == 0) //0 for equal
			return i;
	}
	return 0;
}


/**
	Returns the index of the element with property matching the value
*/
listFindElementByProperty: list, property, value { 
	for(i=1; i<=list.count(); i++) {
		if(list[i][property] == value)
			return i;
	}
	return 0;
}


/**
	Change the property of an element.
	NEW.
*/
listChangeElementProperty: list, index, property, value { 
	if(index <= list.count() && property <= list[index].count()) {
		list[index][property] = value;
	}
	return 0;
}


/**
	Return property value of an element.
	NEW.
*/
listGetElementPropertyValue: list, index, property { 
	if(index <= list.count() && property <= list[index].count()) {
		return list[index][property];
	}
	return nil;
}



/**
	Adds element to list function.
	Returns element index.
	
	bug: for some reason both the last and previous last will be assinged the element after delete in the middle. I figure it must have something to do with the packing and the internal counter.
*/
listAddElement: list, element { 
	list[list.count()+1] = element;
	return list.count();
}


/**
	Same as listAddElement(...), except doesn't add duplicates.
	Requires listCompare(...) for custom comparisons/criteria for equal.
	Rreturns index or 0 if equal element already exists.
*/
listAddUniqueElement: list, element {
	index = listFindElement(list, element);
	if(index > 0)
		return 0;
		
	return listAddElement(list, element);
}


/**
	removes an element from the list.
	returns the removed element if successful, else nil.
	to remove by similar element or property, use find functions to get index.
	be careful to not remove with outdated indices. Each time an element is removed the indices might shift.
*/
listRemoveElementByIndex: list, index {
	if(index<1 || index>list.count())
		return nil;
	
	element = list[index];
	list[index] = nil;
	listPack(list);
	return element;
}

/**
	removes an element from the list by propert value. One element at a time.
	returns the removed element if successful, else nil.
*/
listRemoveElementByProperty: list, property, value{
	index = listFindElementByProperty(list, property, value);
	return listRemoveElementByIndex(list, index);
}

/**
	packs and truncs the list.
	usefull if you want to remove alot of elements by index.
	using listRemoveElementByIndex will pack and trunk after every removal, this might not be desired.
*/
listPack: list {
	list.pack();
	list.trunc(); //we dont want to risk loosing the original allocated memory adress, nor inhibit performance
}

/**
	adds an element to the end of the list.
	returns the index of the element, else 0.
*/
listPush: list, element {
	return listAddElement(list, element);
}

/**
	removes an element from the end of the list.
	returns the removed element if successful, else nil.
*/
listPop: list {
	return listRemoveElementByIndex(list, list.count());
}

/**
	Inserts element at index or last, if index>count.
	offsets succeeding elements.
	returns the index.
*/
listInsert: list, index, element {
	for(i=list.count(); i>=index; i--) { //1 2 3 4 5
		list[i+1] = list[i];
	}
	list[index] = element;
	return index;
}

///
///		LIBLIST - END
///

//internal variables
commandStack;
enabledCommands;
blockedCommands;
lastCommandString; //set in process()

addCommand: list, commandName, commandState {
	return listAddUniqueElement(list, listNewElement(@commandName, commandState@));
}

removeCommand: list, commandName {
	return listRemoveElementByProperty(list, COMMAND_NAME, commandName);
}

findCommand: list, commandName {
	return listFindElementByProperty(list, COMMAND_NAME, commandName);
}

enableCommand: commandName, commandState {
	result = addCommand(enabledCommands, commandName, commandState);
	removeCommand(blockedCommands, commandName); //if blocked unblock
	return result;
}
disableCommand: commandName {
	result = removeCommand(enabledCommands, commandName);
	return result;
}
blockCommand: commandName {
	result = addCommand(blockedCommands, commandName, COMMAND_STATE_BLOCKED);
	removeCommand(enabledCommands, commandName); //if enabled disable
	return result;
}
unblockCommand: commandName {
	result = removeCommand(blockedCommands, commandName);
	return result;
}
setCommandProperty: list, index, property, value {
	result = listChangeElementProperty(list, index, property, value);
	return result;
}
getCommandProperty: list, index, property {
	return listGetElementPropertyValue(list, index, property);
}

validateCommand: commandString {
	if(!commandString || commandString == "") {
		//error("["+command+"] is not a valid command.");
		return false;
	}
	return true;
}

executeCommand: command {
	items = Scene().getSelect();
	
	for(i=1; i<=items.count(); i++) {
		SelectItem(items[i].id);
		CommandInput(command);
	}
	Select(items);
}

executeCommandSkipFirst: command {
	items = Scene().getSelect();

    for(i=2; i<=items.count(); i++) {
        SelectItem(items[i].id);
        CommandInput(command);
    }
    Select(items);
}

executeLastCommand { //executes last command, uses multiservice for commands that are enabled, otherwise not.
	if(!validateCommand(lastCommandString))
		return;

	lastCommand = parse(" ", lastCommandString); //I should have this globally
	isEnabled = findCommand(enabledCommands, lastCommand[1]);
	
	if(executeLastCommandOptions(lastCommandString, isEnabled)) {
		if(isEnabled)
			executeCommand(lastCommandString);
		else
			CommandInput(lastCommandString);
	}
}

//replacement. but popup should be able to show all last commands, not just enabled.
executeLastCommandOptions3: command, isEnabled {	
    reqbegin("Execute Last Command");
	
	ctl_lastCommands = ctlpopup("Execute", 1, "evt_showLastCommands");
	
    answer = reqpost();
	if(answer) {
		answer = getValue(ctl_lastCommands);
	}
    reqend();
	
	return answer;
}

executeLastCommandOptions: command, isEnabled {
	var answer;
	if(!isEnabled) {
		answer = dialog(
		NAME+" - "+"Execute Last Command",
		"Would you like to execute ["+command+"]?",
		"(This command has not been enabled and will be executed without the aid of MultiService.)");
	}
	else {
	answer = dialog(
		NAME+" - "+"Execute Last Command",
		"Would you like to execute ["+command+"] on all selected itdems?",
		"");
	}
	return answer;
}

Select: items {
	if(!items || !items.count()) return;
	
    SelectItem( items[items.count()].id );
    for( i = items.count(); i > 0; i-- )
    AddToSelection( items[i].id );
}

///
///		COMMAND LIST - END
///


///
/// 	INPUT-OUTPUT START
///

defaultFilename = "ef_MultiService.cfg";
exportFilename = "ef_MultiService.cfg";
exportFolder = nil;


LoadConfiguration {
    (a,b,c,d) = split(SCRIPTID); //gets full path of script
	filepath = string(a,b,defaultFilename);
	
    LoadConfiguration13(filepath);

	/*
	//works, but legacy. maybe only allow from import.
    //Try to load old config
    filename = SCRIPTID; //gets full path of script
    filename ~= r~ls.?$~dat~; //exchanges extension .ls or .lsc for .dat
    file = File(filename, "r");
    if(file != nil && file.isOpen() && !file.eof()) {
        LoadConfiguration11(file);
        filedelete(filename);
        SaveConfiguration();
    }
	*/
}

LoadConfiguration13: filepath {
	file = File(filepath, "r");

	if(!file) {
		error("Could not load from ", filepath);
		return;
	}
	
    while(!file.eof()) {
        line = parse(" ", file.read());
        for(i=3; i<=line.count(); i++) { //compounds arguments or multiword commands
            line[2] += " "+line[i];
        }
        
        select
            when line[1] == "Version" then do
                //do nothing
                //line ~= r~Version ~~;
                //version = round(number(line),1);
                end
            when line[1] == "ActiveCommand" then do
				if(validateCommand(line[2])) {
					enableCommand(line[2], COMMAND_STATE_ACTIVE);
				}
                end
            when line[1] == "InactiveCommand" then do
				if(validateCommand(line[2])) {
					enableCommand(line[2], COMMAND_STATE_INACTIVE);
				}
                end
            when line[1] == "BlockedCommand" then do
                if(validateCommand(line[2])) {
					blockCommand(line[2]); //second param ignored
				}
                end
        end
    }
    file.close();
}

LoadConfiguration11: file {
    if(file != nil && file.isOpen()) {
        version = number( file.read() );
        enabledCommandsCnt = integer( file.read() );
        for(i = 1; i <= enabledCommandsCnt && !file.eof(); i++) {
            wasAdded = listAddUniqueElement( enabledCommands, listNewElement(@file.read(), COMMAND_STATE_ACTIVE@) );
        }
    }
    file.close();
}

SaveConfiguration {
    (a,b,c,d) = split(SCRIPTID); //gets full path of script
	filepath = string(a,b,defaultFilename);
	SaveConfiguration13(filepath);
}

SaveConfiguration13: filepath {

    file = File(filepath, "w");
	
	if(!file) {
		error("Could not dave to ", filepath);
		return;
	}

    if(file != nil && file.isOpen()) {
        file.writeln("Version "+VERSION);
        for(i = 1; i <= enabledCommands.count(); i++) {
			if(validateCommand(enabledCommands[i][COMMAND_NAME])) {
				if(enabledCommands[i][COMMAND_STATE])
					file.writeln("ActiveCommand "+enabledCommands[i][COMMAND_NAME]);
				else
					file.writeln("InactiveCommand "+enabledCommands[i][COMMAND_NAME]);
			}
        }
        for(i = 1; i <= blockedCommands.count(); i++) {
			if(validateCommand(blockedCommands[i][COMMAND_NAME])) {
				file.writeln("BlockedCommand "+blockedCommands[i][COMMAND_NAME]);
			}
        }
        //file.writeln("PanelSize "+panelWidth+" "+panelHeight+panelOffset);
        //file.writeln("PluginManagement "+pluginManagement);
        file.close();
    }
}

ExportConfiguration {
	filepath = getfile("Export MultiService Configuration", exportFilename, exportFolder, 0); //0 for save dialog
	
	if(filepath) {
		(drive,path,filename,ext) = split(filepath);
		if(ext==nil) ext = ".cfg";
		if(file==nil) filename = "ef_MultiService";
		exportFolder = string(drive,path);
		exportFilename = string(filename,ext);
		
		filepath = AddExtension(filepath, ".cfg");
		SaveConfiguration13(filepath);
	}
}

ImportConfiguration {
	filepath = getfile("Import MultiService Configuration", exportFilename, exportFolder, 1); //0 for load dialog

	if(filepath) {
		(drive,path,filename,ext) = split(filepath);
		exportFolder = string(drive,path);
		exportFilename = string(filename,ext);
		
		LoadConfiguration13(filepath);
	}
}

AddExtension: filepath, ext { //ext ex ".cfg"
	if(strright(filepath, 4) != ext) {
		filepath = string(filepath,ext);
	}
	return filepath;
}

LoadCommandList { //not used at the moment
	x_trimLine = regexp(" .*$", "");

    (a,b,c,d) = split(SCRIPTID);
    filepath = string(a,b,"Commands.txt"); //matches any filename and substitutes it
    
	file = File(filepath, "r");
	if(!file) {
		error("Commands.txt missing from script directory.");
	}
	
	commandList[file.linecount()] = nil;
	indexedList;
	i=0;
	while(!file.eof()) {
		i++;
		line = file.read();
		line ~= x_trimLine;
		commandList[commandList.count()+1] = line;
		indexedList[line] = i;
	}
	
	res = commandList.indexOf("SoftFilter");
	res = indexedList["SoftFilter"];
	
	//good bit about regular array is that it is ordered. Which is useful for the gui.
	//good bit about indexList is that it is searchable on property.
	
	
    file.close();
}

///
/// 	INPUT-OUTPUT END
///


///
///		INTERFACE START
///

ctl_minmaxPanel;
ctl_currentCommand;
ctl_lastCommands;
ctl_addCommand;
ctl_blockCommand;
ctl_executeCommand;
ctl_tab;
ctl_sep;
ctl_enabledCommands;
ctl_blockedCommands;
ctl_disableCommands;
ctl_activateCommands;
ctl_deactivateCommands;

panelTitle;
panelWidth;panelHeight;
panelWidthMin;panelHeightMin;
panelWidthMax;panelHeightMax;
panelWidthDefault;panelHeightDefault;
panelHeightOffset;

MAX_LASTCMD;	//number of commands stored in the last commands dropdown menu
lastCommandNames; 		//list of the last issued commands
currentCommandName; 	//the last issued command or user entered or picked command 

currentTab;
@sequence {
	TAB_ENABLED_COMMANDS = 1,
	TAB_BLOCKED_COMMANDS
}
@sequence {
	CTL_ENABLED_COMMANDS = 1,
	CTL_BLOCKED_COMMANDS
}


check; blank; divider;
///ICONS START

check_icon = @ 	"..........",
				"..........",				
				".........o",
				"........oo",
				".......oo.",
				"oo....oo..",
				"ooo..oo...",
				".ooooo....",
				"..ooo.....",
				"...o......",
				".........."
			@;
			
blank_icon = @ 	"..........",
				"..........",
				"..........",
				"..........",
				"..........",
				"..........",
				"..........",
				"..........",
				"..........",
				"..........",
				".........."
			@;			
	
divider_icon = @"o",
				"o",
				"o",
				"o",
				"o",
				"o",
				"o",
				"o",
				"o",
				"o",
				"o",
				"o",
				"o",
				"o"
			@;

check = Icon(check_icon);
blank = Icon(blank_icon);
divider = Icon(divider_icon);

///ICONS END



showPanel
{
	if(!panelTitle) {
		panelTitle = "ef MultiService v"+VERSION;

		check = Icon(check_icon);
		blank = Icon(blank_icon);
		divider = Icon(divider_icon);
		
		currentCommandName = "";
		MAX_LASTCMD = 9;

		panelWidthDefault = 308;
		panelWidthMin = panelWidthDefault;
		panelWidthMax = panelWidthDefault;
		panelWidth = panelWidthDefault;

		panelHeightDefault = 474;
		panelHeightMin = 94;
		panelHeightMax = 2048;
		panelHeight = panelHeightDefault;
		panelHeightOffset = 0;
		currentTab = TAB_ENABLED_COMMANDS;
	}

    if( reqisopen() )
        reqend();

    reqbegin(panelTitle);
    reqsize(panelWidth, panelHeight);
	//reqresize("resize",panelWidthMin, panelHeightMin, panelWidthMax, panelHeightMax);
    
	buttonWidth = 144;
	rowOffset = 22;
	column[1] = 7;
	column[2] = 157;
	column[3] = 276;
	
    //CONTROLS
    ctl_currentCommand = ctlstring("Command", currentCommandName);
    ctl_lastCommands = ctlpopup( "", 1, "evt_showLastCommands");
    ctl_addCommand = ctlbutton("Enable Command", buttonWidth, "evt_enableCommand");
	// ctl_executeCommand = ctlbutton("Execute Command", buttonWidth, "evt_executeCurrentCommand");
	ctl_blockCommand = ctlbutton("Block Command", buttonWidth, "evt_blockCommand");
	//ctl_sep = ctlsep(); ctl_sep.active(0);//draw before tabs
    ctl_tab = ctltab("Enabled Commands", "Blocked Commands");
	
	//Tab 1
	ctl_enabledCommands = ctllistbox("A"+"     "+"Command", 272, 280+panelHeightOffset, "evt_enabledCommandCount", "evt_enabledCommandName", "evt_enabledCommandSelect");
	ctl_activateCommands = ctlbutton("Activate Command(s)", buttonWidth, "evt_activateSelectedCommands");
	ctl_deactivateCommands = ctlbutton("Deactivate Command(s)", buttonWidth, "evt_deactivateSelectedCommands");
    ctl_disableCommands = ctlbutton("Remove Command(s)", buttonWidth, "evt_removeSelectedCommands");
	ctl_blockCommands = ctlbutton("Block Command(s)", buttonWidth, "evt_blockSelectedCommands");
	ctlpage(1, ctl_enabledCommands, ctl_activateCommands, ctl_deactivateCommands, ctl_disableCommands, ctl_blockCommands);
	
	//Tab 2
	ctl_blockedCommands = ctllistbox("  "+"     "+"Command", 272, 280+panelHeightOffset, "evt_blockedCommandCount", "evt_blockedCommandName", "evt_blockedCommandSelect");
	ctl_unblockCommands = ctlbutton("Remove Command(s)", buttonWidth, "evt_unblockCommands");
	ctl_enableCommands = ctlbutton("Enable Command(s)", buttonWidth, "evt_enableCommands");
	ctlpage(2, ctl_blockedCommands, ctl_unblockCommands, ctl_enableCommands);

	//bottom
	ctl_bottomSeparator = ctlsep();
	ctl_importConfiguration = ctlbutton("Import Configuration", buttonWidth, "evt_importConfiguration");
	ctl_exportConfiguration = ctlbutton("Export Configuration", buttonWidth, "evt_exportConfiguration");
	
	//panel controls
	//ctl_minmaxPanel = ctlbutton("", 19, "minmaxPanel");
	
	
	
	
	//additional callbacks
	ctlrefresh(ctl_currentCommand, "evt_setCurrentCommand");
    ctlrefresh(ctl_lastCommands, "evt_setCurrentCommand");
	ctlrefresh(ctl_tab, "evt_setTab");
	

    //POSITION
    row = 10;
	//ctlposition(ctl_minmaxPanel, column[1], panelHeight - 12, panelWidth - 14, 9);
    ctlposition(ctl_currentCommand, column[1], row, 273);
    ctlposition(ctl_lastCommands, column[3], row, 25);
    row += rowOffset + 2;
	ctlposition(ctl_addCommand, column[2], row);
	row += rowOffset;
	ctlposition(ctl_blockCommand, column[2], row);
    // ctlposition(ctl_executeCommand, column[2], row);
    // row += 24;
    
    row += rowOffset + 8;
	ctlposition(ctl_tab, column[1], row);
	row += 18;
	// ctlposition(ctl_sep, column[1], row);
	row += 1;
    ctlposition(ctl_enabledCommands, column[1], row);
	ctlposition(ctl_blockedCommands, column[1], row);
    row += ctl_enabledCommands.h + 6;
	ctlposition(ctl_activateCommands, column[1], row);
    ctlposition(ctl_disableCommands, column[2], row);
	ctlposition(ctl_unblockCommands, column[2], row);
	row += rowOffset;
	ctlposition(ctl_deactivateCommands, column[1], row);
	ctlposition(ctl_blockCommands, column[2], row);
	ctlposition(ctl_enableCommands, column[2], row);
	row += rowOffset+4;
	ctlposition(ctl_bottomSeparator, 0, row, panelWidth);
	row += 8;
	ctlposition(ctl_importConfiguration, column[1], row);
	ctlposition(ctl_exportConfiguration, column[2], row);
	
	//row += 24;
	//ctlposition(ctl_sep, column[1], row);
	//row += 12;
	//ctlposition(ctl_pluginManagent, column[1], row);
	
	evt_setTab(currentTab); //opens panel with last set tab
	
    reqopen();
}

hidePanel {
	reqabort();
}

minmaxPanel {
	if(!st_isMinimized)
		panelHeight = panelHeightMin;
	else
		panelHeight = panelHeightDefault;
		
	st_isMinimized = st_isMinimized ? false : true;
	reqsize(panelWidth, panelHeight);
	hidePanel();
	showPanel();
}

//Deactivated. Not properly implemented.
resize: w, h {
	panelWidth = w;
	panelHeight = h;
/*
	offset = h - panelHeight;
	panelHeightOffset += offset;

	//ctl_enabledCommands.visible(0);
	//ctl_activateCommands.visible(0);
	//ctl_deactivateCommands.visible(0);
	//ctl_disableCommands.visible(0);

	ctlposition(ctl_enabledCommands, ctl_enabledCommands.x, ctl_enabledCommands.y, 272, ctl_enabledCommands.h+offset );
	ctl_activateCommands.position( ctl_activateCommands.x, ctl_activateCommands.y+offset );
	ctl_deactivateCommands.position( ctl_deactivateCommands.x, ctl_deactivateCommands.y+offset );
	ctl_disableCommands.position( ctl_disableCommands.x, ctl_disableCommands.y+offset );

	//ctl_enabledCommands.visible(1);
	//ctl_activateCommands.visible(1);
	//ctl_deactivateCommands.visible(1);
	//ctl_disableCommands.visible(1);
	
	panelWidth = w;
	panelHeight = h;
	
	SaveConfiguration();
	*/
}

ReqUpdate {
	//cleanLastCommands();
	e = ctl_enabledCommands.value;
	b = ctl_blockedCommands.value;
	requpdate();
	ctl_enabledCommands.value = e;
	ctl_blockedCommands.value = b;
	
	SaveConfiguration();
}

addToLastCommandNames: commandName {
    if( lastCommandNames && commandName == lastCommandNames[1] ) //dont add same commandName if issued twice in a row
    return;

    for(i=lastCommandNames.count() < MAX_LASTCMD ? lastCommandNames.count(): MAX_LASTCMD; i>=1; i--) {
        lastCommandNames[i+1] = lastCommandNames[i];
    }
	lastCommandNames[1] = commandName;
	evt_setCurrentCommand(lastCommandNames[1]);
}


cleanLastCommands {
    for(i=1; i<=lastCommandNames.count(); i++) {
        if(findCommand(blockedCommands, lastCommandNames[i])) {
			lastCommandNames[i] = nil;
		}
    }
	lastCommandNames.pack();
	lastCommandNames.trunc();
}

evt_showLastCommands {
    return( lastCommandNames );
}


evt_setCurrentCommand: command {
    if(command.isInt())
		currentCommandName = lastCommandNames[getvalue(ctl_lastCommands)]; //from lastCmd popup
    else
		currentCommandName = command; //from layout
    
    if( reqisopen() )
    {
        setvalue( ctl_currentCommand, currentCommandName );
    }
	setdesc(NAME+" :: "+currentCommandName);
}

//DONE
evt_setTab: tab {
	setvalue(ctl_tab, tab);	//sets
	currentTab = getvalue(ctl_tab); //updates internal variable
}

//DONE
evt_enableCommand {

	if(!validateCommand(currentCommandName)) {
		return;
	}

	index = findCommand(enabledCommands, currentCommandName);
	if(!index) {
		index = enableCommand(currentCommandName, COMMAND_STATE_ACTIVE);
	}
	
	ReqUpdate();
	evt_setTab(TAB_ENABLED_COMMANDS);
	indices = @index@;
	ctl_enabledCommands.value = indices;
}




toggleSelectedCommandState {
	indices = getvalue(ctl_enabledCommands);
	for(i=1; i<=indices.count(); i++) {
		state = listGetElementPropertyValue(enabledCommands, indices[i], COMMAND_STATE);
		setCommandProperty(enabledCommands, indices[i], COMMAND_STATE, !state);
	}
    ReqUpdate();
}
evt_activateSelectedCommands {	
	indices = getvalue(ctl_enabledCommands);
	for(i=1; i<=indices.count(); i++) {
		setCommandProperty(enabledCommands, indices[i], COMMAND_STATE, COMMAND_STATE_ACTIVE);
	}
    ReqUpdate();
}
evt_deactivateSelectedCommands {
	indices = getvalue(ctl_enabledCommands);
	for(i=1; i<=indices.count(); i++) {
		setCommandProperty(enabledCommands, indices[i], COMMAND_STATE, COMMAND_STATE_INACTIVE);
	}
    ReqUpdate();
}

evt_removeSelectedCommands {
    indices = getvalue(ctl_enabledCommands);

	if( indices == nil )
		return;

	for(i=1; i<=indices.count(); i++) {
		commandNames += enabledCommands[indices[i]][COMMAND_NAME];
	}

	for(i=1; i<=commandNames.count(); i++) {
		disableCommand(commandNames[i]);
	}

	ReqUpdate();
	ctl_enabledCommands.value = nil;
}
removeCommandOptions: commandNames {
	if(commandNames.count()>1) {
		answer = dialog(
		NAME+" - "+"Remove Command",
		"Are you sure you want to remove the selected commands?",
		"");
	}
	else {
		answer = dialog(
		NAME+" - "+"Remove Command",
		"Are you sure you want to remove ["+commandNames[1]+"]?",
		"");
	}
	return answer;
}
//blocks commands seleted in enabled commands.
evt_blockSelectedCommands {
    indices = getvalue(ctl_enabledCommands);

	if( indices == nil )
		return;

	for(i=1; i<=indices.count(); i++) {
		commandNames += enabledCommands[indices[i]][COMMAND_NAME];
	}
	
	indices = nil;
	for(i=1; i<=commandNames.count(); i++) {
		indices += blockCommand(commandNames[i]);
	}
	
	ReqUpdate();
	evt_setTab(TAB_BLOCKED_COMMANDS);
	ctl_blockedCommands.value = indices;
}



evt_importConfiguration {
	ImportConfiguration();
	ReqUpdate();
}
evt_exportConfiguration {
	ExportConfiguration();
}





evt_enabledCommandCount {
    return( enabledCommands.count() );
}
evt_enabledCommandName: i {
	icon = enabledCommands[i][COMMAND_STATE] ? check : blank;
    return( ""+icon+"     "+enabledCommands[i][COMMAND_NAME] );
}
evt_enabledCommandSelect: items { //selecting multiple rows, generates an event for every new row.
	if(index = wasDoubleClicked(CTL_ENABLED_COMMANDS)) {
		toggleSelectedCommandState();
	}
}



//DONE
evt_blockCommand {

	if(!validateCommand(currentCommandName)) {
		return;
	}
	
	index = findCommand(blockedCommands, currentCommandName);
	if(!index) {
		index = blockCommand(currentCommandName);
	}
	
	ReqUpdate();
	evt_setTab(TAB_BLOCKED_COMMANDS);
	indices = @index@;
	ctl_blockedCommands.value = indices;
}
evt_blockCommandOptions: command {
	answer = dialog(
		NAME+" - "+"Block Command",
		"Are you sure you want to block ["+command+"]?",
		"(To unblock you will have to edit the configuration file.)");
	return answer;
}
evt_unblockCommands {

    indices = ctl_blockedCommands.value;

	if( indices == nil )
		return;

	for(i=1; i<=indices.count(); i++) {
		commandNames += blockedCommands[indices[i]][COMMAND_NAME];
	}

	for(i=1; i<=commandNames.count(); i++) {
		unblockCommand(commandNames[i]);
	}
	
	ReqUpdate();
	ctl_blockedCommands.value = nil;
}

//enables commands selected in blocked commands
evt_enableCommands {

    indices = ctl_blockedCommands.value;

	if( indices == nil )
		return;

	for(i=1; i<=indices.count(); i++) {
		commandNames += blockedCommands[indices[i]][COMMAND_NAME];
	}

	indices = nil;
	for(i=1; i<=commandNames.count(); i++) {
		indices += enableCommand(commandNames[i], COMMAND_STATE_ACTIVE);
	}
	
	ReqUpdate();
	evt_setTab(TAB_ENABLED_COMMANDS);
	ctl_enabledCommands.value = indices;
}



evt_blockedCommandCount {
    return( blockedCommands.count() );
}
evt_blockedCommandName: i {
	//icon = blockedCommands[i][COMMAND_STATE] ? check : blank;
    return( " "+"  "+"     "+blockedCommands[i][COMMAND_NAME] );
}
evt_blockedCommandSelect: items { //selecting multiple rows, generates an event for every new row.
}




wasDoubleClicked: proxy_id { //uses user specified id in lieu of non existant ctl.id

	if(!st_last_type) {
		st_last_type = 0;
		st_last_index = 0;
		st_last_tick = 0;
	}

	(,,,tick) = time();

	was_double_clicked=0;
	indices=nil;
	switch(proxy_id) {
		case CTL_ENABLED_COMMANDS: indices=ctl_enabledCommands.value; break;
		case CTL_BLOCKED_COMMANDS: indices=ctl_blockedCommands.value; break;
	}
	index = indices.count() == 1 ? indices[1] : 0;
	if(!index) return 0; //Early exit! no nil, no multiple selection.
	
	same_type = proxy_id==st_last_type;
	same_index = index==st_last_index;
	in_time = (tick-st_last_tick)<1;
	if(same_type && same_index && in_time) {
		was_double_clicked = 1;
		tick = 0; //reset counter so third immediate click doesnt generate second doubleclick
	}
	st_last_type  = proxy_id;
	st_last_index = index;
	st_last_tick  = tick;

	//info(was_double_clicked, "-",same_type,"-",same_index,"-",in_time);
	if(was_double_clicked)
		return index;
	else
		return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////
/// HELPER FUNCTIONS /////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

dialog: title, text1, text2 {
	//close main panel
	reqWasOpen = 0;
	if(reqisopen()) {
		reqWasOpen = 1;
		reqabort();
	}
	
    reqbegin(title);
	if( text1 != nil && text1 != "" )
		ctltext("", text1);
	if( text2 != nil && text2 != "" )
		ctltext("", text2);
    answer = reqpost();
    reqend();
	
	//reopen main panel
	if(reqWasOpen)
		showPanel();	
	
	return answer;
}

///
///		INTERFACE END
///



///
///		INSTANCE - START
///

//requires SERVERNAME
@define INSTANCE_RING_NAME SERVERNAME+"_Instances"
@define INSTANCE_SERVERNAME SERVERNAME
@define INSTANCE_SERVERTYPE "MasterHandler"
@define INSTANCE_SERVERITEM_FUNC Scene
///LIBSERVER START

/**
	Lib to handle instances for persistant Layout scripts.
	
	Currently only has means to prevent multiple instances.
	It accomplishes this by having all instance broadcast
	their creation. This broadcast will only be received
	by a previous instance and it can then trigger a
	RemoveServer operation to remove the new instance.
	
	Removal must be done in the process function since LW
	hasn't added the server yet when the comring message is sent.
	
	To use:
		-define INSTANCE_RING_NAME
		the name of the ring. should be unique so as not to
		interfere with other plugins.
		-define INSTANCE_SERVERNAME
		-define INSTANCE_SERVERTYPE
		-define INSTANCE_SERVERITEM_FUNC
		name of function returning oa the server is attached to
*/
@sequence {
	INSTANCE_ATTACH,
	INSTANCE_DETACH
}
registerInstance { //only sent by second instance
	comringattach(INSTANCE_RING_NAME,"instanceRingEvent");
	comringmsg(INSTANCE_RING_NAME,INSTANCE_ATTACH);
}
unregisterInstance { //only sent by second instance
	comringdetach(INSTANCE_RING_NAME);
}
doRemove = false;
instanceRingEvent: eventCode, data { //only processed by first instance
	if(eventCode==INSTANCE_ATTACH)
		doRemove = true; //delay action; server hasn't added yet.
}
removeCompetingInstances: type, name, item {
	servers = item.server(type);
	for(i=1, hasSkippedFirst=false, wasRemoved=0; i<=servers.count(); i++) {
		if(servers[i] == name) {
			if(!hasSkippedFirst) {
				hasSkippedFirst = true;
			}
			else {
				RemoveServer(type, i);
				wasRemoved++;
			}
		}
	}
	return wasRemoved;
}
maintainSingleInstance {
	if(doRemove) {
		wasRemoved = removeCompetingInstances(INSTANCE_SERVERTYPE,INSTANCE_SERVERNAME,INSTANCE_SERVERITEM_FUNC());
		if(wasRemoved)
			pluginErrorOptions();
		doRemove = false;
	}
}
pluginErrorOptions {
	warn("An instance of "+INSTANCE_SERVERNAME+" already exists. ",
		 "Only one is instance allowed.");
}
///LIBSERVER END


///
///		INSTANCE - END
///


/// 	EVENTS START
event_PluginChanged {
	ParentInPlace();
}
/// 	EVENTS END

@define EXECUTE_LAST_COMMAND "Generic_ef_MultiService_LastCommand" //the commandString that triggers executeLastCommand function.
@define COMMAND_LIST_LENGTH 100
@define COMMAND_STACK_LENGTH 10



create {
	setdesc(NAME);

	enabledCommands = listNew(COMMAND_LIST_LENGTH);
	blockedCommands = listNew(COMMAND_LIST_LENGTH);
	commandStack = listNew(COMMAND_STACK_LENGTH);
    LoadConfiguration();
	//LoadCommandList();
	
	//registerInstance();
}

destroy {
	//unregisterInstance();
}

load: what,io
{
}

save: what,io
{
}

flags {
	//mapevent(LWEVNT_NOTIFY_PLUGIN_CHANGED,"event_PluginChanged"); //doesnt work, likely plugin events arent implemented properly.
    return( LAYOUT );
}

options {
	showPanel();
}

process: event, command {
	//crashes on render at one time! Dunno why.
	if( event == RENDER_DONE )
		return;

    if( event == COMMAND ) {
		//maintainSingleInstance(); //on refresh()
		
        commandLine = parse(" ", command);
		commandString = commandLine[1] + ((commandLine[2] == "(null)" || commandLine[2] == "(Null)") ? "" : " " + commandLine[2]); //cleaned command string
	
        isEnabled = listFindElementByProperty(enabledCommands, COMMAND_NAME, commandLine[1]); //returns index if found
		isActive = isEnabled ? enabledCommands[isEnabled][COMMAND_STATE] : 0;
		isBlocked = listFindElementByProperty(blockedCommands, COMMAND_NAME, commandLine[1]);

		//some commands just give error message on multiselect.
		if(commandLine[1] == "Generic_BT_RecPivRot") {
			doFirst = true;
		}
		//execute as last command
		if(commandString == EXECUTE_LAST_COMMAND) {
			executeLastCommand();
			return;
		}
		//or as enabled command
        else if( isEnabled && isActive ) {
		
			if(doFirst) { 
				al = Scene().alertlevel;
				AlertLevel(ALERT_EXPERT);
				executeCommand(commandString);
				info("");//hide error messages, except first, whihc we cant get to. (for BT_RecPivRot)
				AlertLevel(al);
				doFirst = false;
			}
            else executeCommandSkipFirst(commandString); 
        }

		//if the command isn't blocked add it to the list of last executed commands.
        if( !isBlocked ) {
			addToLastCommandNames( commandLine[1] );
			lastCommandString = commandString;
		}
    }
}


